
//Quick WebGL provides simple and easy functions for doing shader processing behind the scenes(rendering to fullscreen quad)
var gl,quickGLCanvas;



//Helper function for setting texture filtering options.
function SetTextureOptions(linear,mipmap,repeatx,repeaty) {
	var mgf, mif;
	if (linear) {
		mgf = gl.LINEAR;
		mif = mipmap?gl.LINEAR_MIPMAP_LINEAR:gl.LINEAR;
	} else {
		mgf = gl.NEAREST;
		mif = mipmap?gl.NEAREST_MIPMAP_LINEAR:gl.NEAREST;
	}
	gl.texParameteri(gl.TEXTURE_2D,gl.TEXTURE_MAG_FILTER,mgf);
	gl.texParameteri(gl.TEXTURE_2D,gl.TEXTURE_MIN_FILTER,mif);
	gl.texParameteri(gl.TEXTURE_2D,gl.TEXTURE_WRAP_S,repeatx?gl.REPEAT:gl.CLAMP_TO_EDGE);
	gl.texParameteri(gl.TEXTURE_2D,gl.TEXTURE_WRAP_T,repeaty?gl.REPEAT:gl.CLAMP_TO_EDGE);
}


var boundTextures = [];

//Bind texture tex to uniform location uni.
function BindTexture(uni,tex) {
	var ni = boundTextures.indexOf(gl.INVALID_VALUE);
	if (ni === -1) ni = boundTextures.push(uni)-1;
	else boundTextures[ni] = uni;
	gl.activeTexture(gl.TEXTURE0+ni);
	gl.bindTexture(gl.TEXTURE_2D,tex);
	gl.uniform1i(uni,ni);
}

//Unbind texture at uniform location uni.
function UnbindTexture(uni) {
	boundTextures[boundTextures.indexOf(uni)] = gl.INVALID_VALUE;
}
//Unbind all textures.
function UnbindAllTextures() {
	for (var i = 0; i < boundTextures.length; i++) boundTextures[i] = gl.INVALID_VALUE;
}


var shaders = [], shaderGlobals = {}, currentShader = null;

//Helper function for compiling vertex/fragment webgl shader.
function CompileShader(src,type) {
	var s = gl.createShader(type);
	gl.shaderSource(s,src);
	gl.compileShader(s);
	if (!gl.getShaderParameter(s,gl.COMPILE_STATUS)) {
		console.error("Error compiling GLSL shader: "+gl.getShaderInfoLog(s)+"\nFrom shader:");
		console.log(src);
		gl.deleteShader(s);
		return null;
	}
	return s;
}

/**Create new WebGL vertex/fragment shader from src code string.
@constructor*/
function Shader(src) {
	var inci = 0;
	while (inci < src.length && (inci = src.indexOf('#include "',inci)) !== -1) {
		var endi = src.indexOf('"',inci+10);
		if (endi !== -1) {
			var id = document.getElementById(src.substring(inci+10,endi));
			if (id) src = src.substring(0,inci)+id.textContent+src.substring(endi+1);
			inci = endi;
		} else inci++;
	}
	src = "#version 100\nprecision highp float;precision highp int;"+src;
	
	var vertStr = "void VERT()",
	fragStr = "void FRAG()",
	mainStr = "void main()",
	vertInd = src.indexOf(vertStr),
	fragInd = src.indexOf(fragStr);
	if (vertInd === -1 || fragInd === -1) return;
	
	var fragCode = src.substring(0,vertInd)+mainStr+src.substring(fragInd+11);
	inci = 0;
	while ((inci = fragCode.indexOf("attribute",inci)) !== -1) {
		var endi = fragCode.indexOf(";",inci+9);
		if (endi !== -1) {
			fragCode = fragCode.substring(0,inci)+fragCode.substring(endi+1);
		} else inci++;
	}
	
	var vert = CompileShader(src.substring(0,fragInd).replace(vertStr,mainStr),gl.VERTEX_SHADER),
	frag = CompileShader(fragCode,gl.FRAGMENT_SHADER);
	if (!vert || !frag) return;
	
	var sp = gl.createProgram();
	gl.attachShader(sp,vert);
	gl.attachShader(sp,frag);
	gl.bindAttribLocation(sp,0,"vpos");
	gl.linkProgram(sp);
	gl.detachShader(sp,vert);
	gl.detachShader(sp,frag);
	gl.deleteShader(vert);
	gl.deleteShader(frag);	
	if (!gl.getProgramParameter(sp,gl.LINK_STATUS)) {
		console.error("Error linking GLSL program: "+gl.getProgramInfoLog(sp));
		return;
	}
	
	var nuni = gl.getProgramParameter(sp, gl.ACTIVE_UNIFORMS);
	for (var i = 0; i < nuni; i++) {
		var u = gl.getActiveUniform(sp,i),
			as = u.name.indexOf("[");
		this[as===-1?u.name:u.name.substring(0,as)] = gl.getUniformLocation(sp,u.name);
	}
	this.globalsStateTime = 0;
	this.program = sp;
	shaders.push(this);
}

//Bind shader for rendering.
Shader.prototype.Use = function() {
	if (currentShader === this) return;
	if (currentShader) {
		for (var k in shaderGlobals) {
			if (k in currentShader) {
				if (!shaderGlobals[k].data) UnbindTexture(this[k]);
			}
		}
	}
	currentShader = this;
	gl.useProgram(this.program);
	for (var k in shaderGlobals) {
		var ul = this[k];
		if (ul !== undefined) {
			var sg = shaderGlobals[k];
			if (sg.data) {
				if (sg.stateTime > this.globalsStateTime) gl.uniform4fv(ul,sg.data);
			} else BindTexture(ul,sg.texture);
		}
	}
	this.globalsStateTime = Date.now();
}

//Delete shader freeing memory.
Shader.prototype.Delete = function() {
	gl.deleteProgram(this.program);
	shaders.splice(shaders.indexOf(this),1);
	if (currentShader === this) {
		for (var k in shaderGlobals) {
			if (k in this) {
				if (!shaderGlobals[k].data) UnbindTexture(this[k]);
			}
		}
		currentShader = null;
	}
}



/**Create new shader global, uniform data accessible from any shader, if dataSz is 0 data is a texture otherwise data type is Float32Array.
@constructor*/
function ShaderGlobal(n,dataSz) {
	this.name = n;
	this.stateTime = 0;
	if (dataSz) {
		this.data = new Float32Array(dataSz);
		this.texture = null;
	} else {
		this.data = null;
		this.texture = null;
	}
	shaderGlobals[n] = this;
}

//Update shader global data.
ShaderGlobal.prototype.Update = function() {
	this.stateTime = Date.now();
	if (currentShader) {
		var ul = currentShader[this.name];
		if (ul !== undefined) {
			if (this.data) gl.uniform4fv(ul,this.data);
			else {
				UnbindTexture(ul);
				BindTexture(ul,this.texture);
			}
		}
	}
}



//Rendertexture class, easy to use framebuffer object interface
var currentRenderTexture = null, display = {framebuffer:null,width:0,height:0},
currentRenderTextureX, currentRenderTextureY, currentRenderTextureW, currentRenderTextureH;

/**Create new webgl rendertexture(framebuffer) w width and h height in pixels
@constructor*/
function RenderTexture(w,h) {
	this.framebuffer = gl.createFramebuffer();
	this.texture = gl.createTexture();
	this.width = w;
	this.height = h;
	var twh = w+h;
	this.aspectX = w/twh;
	this.aspectY = h/twh;
	gl.bindTexture(gl.TEXTURE_2D,this.texture);
	gl.texImage2D(gl.TEXTURE_2D,0,gl.RGBA,w,h,0,gl.RGBA,gl.UNSIGNED_BYTE,null);
	SetTextureOptions(false,false,false,false);
	gl.bindFramebuffer(gl.FRAMEBUFFER,this.framebuffer);
	gl.framebufferTexture2D(gl.FRAMEBUFFER,gl.COLOR_ATTACHMENT0,gl.TEXTURE_2D,this.texture,0);
	currentRenderTexture = this;
	currentRenderTextureX = currentRenderTextureY = 0;
	currentRenderTextureW = w; currentRenderTextureH = h;
}

//Resize render texture.
RenderTexture.prototype.SetSize = function(w,h,internalFormat,format,type,dat) {
	if (!internalFormat) internalFormat = gl.RGBA;
	if (!format) format = gl.RGBA;
	if (!type) type = gl.UNSIGNED_BYTE;
	this.width = w;
	this.height = h;
	var twh = w+h;
	this.aspectX = w/twh;
	this.aspectY = h/twh;
	
	gl.bindTexture(gl.TEXTURE_2D,this.texture);
	gl.texImage2D(gl.TEXTURE_2D,0,internalFormat,w,h,0,format,type,dat);
}

//Bind sub window(x,y,w,h) of rendertexture as the current rendertarget.
RenderTexture.prototype.Bind = function(x,y,w,h) {
	if (arguments.length < 4) {
		x = y = 0;
		w = this.width; h = this.height;
	}
	if (currentRenderTexture === this && currentRenderTextureX === x && currentRenderTextureY === y && currentRenderTextureW === w && currentRenderTextureH === h) return;
	currentRenderTexture = this;
	currentRenderTextureX = x; currentRenderTextureY = y;
	currentRenderTextureW = w; currentRenderTextureH = h;
	
	gl.bindFramebuffer(gl.FRAMEBUFFER,this.framebuffer);
	gl.viewport(x,y,w,h);
	gl.scissor(x,y,w,h);
}
display.Bind = RenderTexture.prototype.Bind;

//Delete rendertexture freeing data.
RenderTexture.prototype.Delete = function() {
	gl.deleteFramebuffer(this.framebuffer);
	gl.deleteTexture(this.texture);
}



//create mesh from array of vertex positions of size dimensions(max 4), returns webgl array buffer
function CreateMesh(dat,size) {
	var buf = gl.createBuffer();
	gl.bindBuffer(gl.ARRAY_BUFFER,buf);
	gl.bufferData(gl.ARRAY_BUFFER,new Float32Array(dat),gl.STATIC_DRAW);
	gl.vertexAttribPointer(0,size,gl.FLOAT,false,size*4,0);
	gl.enableVertexAttribArray(0);
	buf.attPtrSize = size;
	return buf;
}
function BindMesh(buf) {
	gl.bindBuffer(gl.ARRAY_BUFFER,buf);
	gl.vertexAttribPointer(0,buf.attPtrSize,gl.FLOAT,false,buf.attPtrSize*4,0);
}



//Initialize quick WebGL canvas and gl context, returns false if failed
var fullscreenTriangle;
function InitializeQuickWebGL(canvas) {
	quickGLCanvas = canvas;
	quickGLCanvas.width = quickGLCanvas.height = 1;
	
	//create webgl context
	var gopt = {alpha:true,antialias:false,depth:false,failIfMajorPerformanceCaveat:false,preserveDrawingBuffer:false};
	gl = quickGLCanvas.getContext("webgl",gopt);
	if (!gl) gl = quickGLCanvas.getContext("experimental-webgl",gopt);
	if (!gl) return false;
	
	gl.enable(gl.SCISSOR_TEST);
	
	//load fullscreen triangle mesh for rendering
	fullscreenTriangle = CreateMesh([2,1, -2,1, 0,-3],2);
	
	return true;
}
