<script type="text/plain" id="Pumpkin.glsl">
//path trace pumpkin scene shader
attribute vec3 vpos;
varying vec4 uv;
uniform sampler2D buffer, background, pumpkin;
uniform vec4 params[10];/*0.xy = aspect/fov
0.z = frame id/1000
0.w = aa pixel size

1.xyz = camera position
1.w = pumpkin height scale
2.xy = camera roll cos/sin
2.zw = camera pitch cos/sin
3.xy = camera yaw cos/sin
3.z = floor height
3.w = floor scale

4.xyz = background tint
4.w = pumpkin dents

5.x = pumpkin main grooves frequency
5.y = pumpkin main grooves scale
5.zw = pumpkin grooves 1 frequency/scale
6.xy = pumpkin grooves 2
6.zw = pumpkin grooves 3

7.xyz = pumpkin color
7.w = floor height

8.xyz = floor tint
8.w pumpkin bounding radius pow 2

9.xyz = light color
9.w = light radius
*/

#define FLOOR_ENABLED 1
#define TRANSPARENT_ENABLED 1

#ifdef FLOOR_ENABLED
#define ITER 100
#else
#define ITER 50
#endif
#define EPSILON 0.02
#define NEPS 0.01
#define SEPS .1

#define PI 3.141592653589793


#ifdef FLOOR_ENABLED
uniform sampler2D floorCol, floorHeight, floorRough;
#endif


//2d rotation matrix
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

//returns pseudo random value 0-1
vec4 hash(vec4 s) {
	s = abs(s);
	vec4 r = fract(s.zxwy*vec4(2334.2378764,1263.237653,862.12378645,5226.827364))*576.987654321;
	return fract(r*mat4(1,-.5,-.25,.75,
				  -.25,-1,.75,.5,
				  -.75,.25,1,.5,
				  .25,-.5,.75,1));
}

//repeating triangle wave
float twave(float x) {
	return abs(fract(x)*2.-1.);
}


//returns vec2(distance,material id)
vec2 geometry(vec3 p) {
	vec2 geo;

	//pumpkin
	vec3 apy = p;
	apy.y *= params[1].w;
	float aplen = length(apy);
	//geo.x = abs(aplen-.95);
	if (aplen < 1.2) {//only do complex details within bounds
		float pang = atan(p.x,p.z)+PI*4.+cos(p.y*15.)*.01+sin(p.y*44.)*0.003,
			plen = length(p.xz),
			grooves = cos(pang*params[5].z)*params[5].w+cos(pang*params[6].x)*params[6].y+cos(pang*params[6].z)*params[6].w+
					pow(max(twave(pang*params[5].x)-.5, 0.)*2., 4.)*params[5].y;

		float pthick = texture2D(pumpkin,vec2(atan(apy.z,apy.x)/6.2831853071+.5,acos(apy.y/aplen)/3.14159265)).x*.05,
			dents = pow(min(.75,abs(apy.y))*max(0.,1.-plen),2.)*params[4].w;//bottom and top dents
		aplen += grooves+dents;
	
		geo.x = abs(aplen-.95);
		if (geo.x <= .05+SEPS) {
			if (pthick < 0.0002) geo = vec2(SEPS,3);
			else geo = vec2(min(SEPS,abs(aplen-.9-pthick)-pthick), pthick<.0495||aplen<.99?4.+aplen:0.);
		} else geo = vec2(geo.x-.05,0);
		
		//stem
		float sd = max(length(p.xz)-.05-grooves,abs(dot(p-vec3(0,.8,0),normalize(vec3(.6,1,-.3))))-.2)-clamp(.05-geo.x*.5, 0.,.05);
		if (sd < geo.x) geo = vec2(sd,2);
		
	} else {
		geo = vec2(aplen-1.,0);
	}
	
	//light
	if (aplen < .95) {
		float lightd = length(p+vec3(0,params[7].w-params[9].w-.1,0))-params[9].w;
		if (lightd < geo.x) geo = vec2(lightd,3);
	}

	//floor
	#ifdef FLOOR_ENABLED
	float fd = p.y+params[7].w;
	if (fd < .2) fd += (1.-texture2D(floorHeight,p.xz*params[3].w).x)*params[3].z;//only apply detail map within bounds
	if (fd < geo.x) geo = vec2(fd,1);
	#endif
	
	return geo;
}

//returns material mat3(vec3 color, float roughness, float emission, float subsurfaceOpacity, float subsurfaceScattering)
mat3 material(vec3 p, float mat, vec4 rand) {
	#ifdef FLOOR_ENABLED
	if (mat == 1.) {//floor
		vec2 puv = p.xz*params[3].w;
		return mat3(texture2D(floorCol,puv).xyz*params[8].xyz, pow(texture2D(floorRough,puv).x,3.),0,0, 0,0,0);
	}
	#endif

	//stem
	if (mat == 2.) {
		float ang = atan(p.x,p.z);
		return mat3(mix(vec3(.0011,.004,.0011),vec3(.59,.42,.27), cos(ang*12.2764+5.*cos(ang+p.y*8.))*clamp(length(p.xz)*10.-.5,0.,.5)+.5),
					.3,0,0, 0,0,0);
	}
	
	//light
	if (mat == 3.) {
		float ll = length(params[9].xyz);
		return mat3(params[9].xyz*(1./max(1e-6,ll)), 1,ll,0, 0,0,0);
	}
	
	//inner pumpkin
	if (mat >= 4.) {
		//gloss
		mat3 om = mat3(vec3(0.982,0.781,0.41), .3,0,max(0.,mat-4.9)*2., .7,0,0);
		if (rand.w < 0.3) {
			om[0].xyz = vec3(.95);
			om[1].x = .005;
		}
		return om;
	}
	
	//pumpkin outside
	if (rand.w < 0.15) return mat3(mix(params[7].xyz,vec3(1),0.5), .025,0,0, 0,0,0);//gloss
	return mat3(params[7].xyz, .25,0,0, 0,0,0);
}

//normal at point
vec3 normal(vec3 p) {
	vec3 n = vec3(geometry(p+vec3(NEPS,0,0)).x,geometry(p+vec3(0,NEPS,0)).x,geometry(p+vec3(0,0,NEPS)).x)-geometry(p).x;
	return (n.x==0.&&n.y==0.&&n.z==0.)?vec3(0,1,0):normalize(n);
}


//raytrace bounds, returns 1e8 if nothing hit
float boundsTrace(vec3 rp, vec3 rd) {
	float pumpBoundsL = dot(rp,rp)-params[8].w,
		pumpBoundsA = dot(rp,rd);
	if (pumpBoundsL <= 0.) return -1.;
	pumpBoundsL = pumpBoundsA*pumpBoundsA-pumpBoundsL;
	if (pumpBoundsL >= 0.) {
		pumpBoundsL = sqrt(pumpBoundsL);
		if (pumpBoundsL < -pumpBoundsA) pumpBoundsL = -pumpBoundsA-pumpBoundsL+EPSILON;
		else pumpBoundsL = 1e8;
	} else pumpBoundsL = 1e8;

	#ifdef FLOOR_ENABLED
	float fbDst = rp.y+params[7].w-EPSILON*2.;
	if (fbDst <= 0.) return -1.;
	if (rd.y < 0.) pumpBoundsL = min(pumpBoundsL,-fbDst/rd.y+EPSILON);
	#endif
	
	return pumpBoundsL;
}


void VERT() {
	uv = vec4(vpos.xy*.5+.5, vpos.xy*params[0].xy+(hash(params[0].z*vec4(.1,1,10,100)).xy-.5)*params[0].w);
	gl_Position = vec4(vpos,1);
}
void FRAG() {
	//setup ray to trace
	vec3 rp = params[1].xyz,
		rd = normalize(vec3(uv.zw,1)),
		c = vec3(1),
		l = vec3(0);
	vec4 random = hash(vec4(uv.yzw,params[0].z));
	bool bounced = false;
	float ldst = 0.;
	
	//camera rotation
	rd.xy *= mat2(params[2].xy,-params[2].y,params[2].x);//roll
	rd.yz *= mat2(params[2].zw,-params[2].w,params[2].z);//pitch
	rd.xz *= mat2(params[3].xy,-params[3].y,params[3].x);//yaw
	
	//initial bounds ray trace
	float boundsDst = boundsTrace(rp,rd);
	
	if (boundsDst < 1e8) {
		rp += rd*max(0.,boundsDst);
		
		//path tracing ray march loop
		for (int i = 0; i < ITER; i++) {
			vec2 geo = geometry(rp);
			if (geo.x < EPSILON) {
				mat3 mat = material(rp,geo.y,random);//sample material
				c *= mat[0];
				l += c*mat[1].y;//emission
			
				random = hash(vec4(rp,params[0].z));
				if (mat[1].z > random.z) {
					rd = normalize(rd+(random.xyw-.5)*mat[2].x);
					geo.x = SEPS;//subsurface scattering
				} else {
					rp -= rd*ldst;
					vec3 nrm = normal(rp),
						tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)),
						bitang = cross(tang,nrm);
					float cs = pow(random.x,mat[1].x), sn = sqrt(1.-cs*cs), ang = random.y*PI*2.;
					rd = reflect(rd,normalize(mat3(nrm,tang,bitang)*vec3(cs,sn*cos(ang),sn*sin(ang))));//opaque scatter, reflect ray off random microsurface bump
					geo.x = EPSILON;
				}
				bounced = true;
			}
			
			#ifdef FLOOR_ENABLED
			float rayDst = boundsTrace(rp,rd);
			if (rayDst < 1e8 && rayDst > -1.) {
				rp += rd*rayDst;
			} else {
				rp += rd*geo.x*(.5+.5*random.z);
				ldst = geo.x;
			}
			#else
			rp += rd*geo.x*(.5+.5*random.z);
			ldst = geo.x;
			#endif
		}

	}
	
	//background
	vec4 bg = pow(texture2D(background,vec2(atan(rd.z,rd.x)/6.2831853071+.5,acos(rd.y)/3.14159265)),vec4(2.2))*params[4];
	#ifdef TRANSPARENT_ENABLED
	if (bounced) l += c*bg.xyz;
	
	//add sample to buffer
	gl_FragColor = (params[0].z==0.?vec4(0):texture2D(buffer,uv.xy))+vec4(l,bounced?1.:0.);
	
	#else
	l += bounced?c*bg.xyz:bg.xyz;
	
	//add sample to buffer
	gl_FragColor = (params[0].z==0.?vec4(0):texture2D(buffer,uv.xy))+vec4(l,1);
	#endif
}
</script>

<script type="text/plain" id="Display.glsl">
//display path tracer
attribute vec3 vpos;
varying vec2 uv;
uniform sampler2D buffer;

void VERT() {
	uv = vpos.xy*.5+.5;
	gl_Position = vec4(vpos,1);
}
void FRAG() {
	vec4 sum = texture2D(buffer,uv);
	gl_FragColor = pow(sum/sum.w,vec4(1./2.2));//normalize and gamma correct
}
</script>

<script type="text/plain" id="DisplayTransparent.glsl">
//display transparent path tracer
attribute vec3 vpos;
varying vec2 uv;
uniform sampler2D buffer;
uniform float samples;

void VERT() {
	uv = vpos.xy*.5+.5;
	gl_Position = vec4(vpos,1);
}
void FRAG() {
	gl_FragColor = pow(clamp(texture2D(buffer,uv)/samples,0.,1.),vec4(1./2.2));//normalize and gamma correct
}
</script>


<script type="text/plain" id="Scoop.glsl">
//scoop pumpkin
attribute vec3 vpos;
varying vec2 uv;
uniform sampler2D map;
uniform vec4 params[2];/*0.xy = 2nd last point, 1e8 if disabled
0.zw = last point
1.xy = current point

1.z = radius
1.w = opacity*/

float sline(vec2 a, vec2 b, float sx) {
	vec2 pa = uv-a, ba = b-a;
	pa -= ba*clamp(dot(pa,ba)/dot(ba,ba),0.,1.);
	pa.x *= sx;
	return length(pa);
}

float scoop(vec2 last, vec2 curr, float sideX, float scaleCorrect) {//scoop brush, capsule line with sine distribution
	float psx = curr.x<.5?-1.:1.;
	vec2 diff = last-curr;
	if (length(diff) < length(diff+vec2(psx,0))) psx = 0.;
	return sin(max(0.,1.-min(sline(last+vec2(psx,0),curr,scaleCorrect),
			sline(last+vec2(psx+sideX,0),curr+vec2(sideX,0),scaleCorrect))/params[1].z)*1.5707963267);
}

void VERT() {
	uv = vpos.xy*.5+.5;
	gl_Position = vec4(vpos,1);
}
void FRAG() {
	float val = texture2D(map,uv).x,
		sideX = uv.x<.5?-1.:1.,
		scaleCorrect = 2.-pow(abs(uv.y-.5)*2.,2.)*1.9,
		sub = scoop(params[0].zw,params[1].xy,sideX,scaleCorrect);//current stroke
	if (params[0].x < 1e8) sub = max(0.,sub-scoop(params[0].xy,params[0].zw,sideX,scaleCorrect));//subtract last stroke
	val -= sub*params[1].w;
	gl_FragColor = vec4(max(val,0.),0,0,1);
}
</script>


<script type="text/plain" id="Knife.glsl">
//knife carve pumpkin
attribute vec3 vpos;
varying vec2 uv;
uniform sampler2D map;
uniform vec4 params[2];/*0.xy = 2nd last point, 1e8 if disabled
0.zw = last point
1.xy = current point

1.z = radius
1.w = opacity*/

float line(vec2 a, vec2 b, float sx) {
	vec2 dir = b-a;
	float dlen = length(dir);
	if (dlen == 0.) {
		return 1e8;
	}
	
	dir *= 1./dlen;
	vec2 rp = uv-b;
	rp.x *= sx;
	float ax = abs(dot(rp,dir)-dlen*.5)-dlen*.5,
		ay = abs(dot(rp,vec2(-dir.y,dir.x)));
	return max(ax,ay);
}

float knife(vec2 last, vec2 curr, float sideX, float scaleCorrect) {//scoop brush, capsule line with sine distribution
	float psx = curr.x<.5?-1.:1.;
	vec2 diff = last-curr;
	if (length(diff) < length(diff+vec2(psx,0))) psx = 0.;
	return max(0.,1.-min(line(last+vec2(psx,0),curr,scaleCorrect),
			line(last+vec2(psx+sideX,0),curr+vec2(sideX,0),scaleCorrect))/params[1].z);
}

void VERT() {
	uv = vpos.xy*.5+.5;
	gl_Position = vec4(vpos,1);
}
void FRAG() {
	float val = texture2D(map,uv).x,
		sideX = uv.x<.5?-1.:1.,
		scaleCorrect = 2.-pow(abs(uv.y-.5)*2.,2.)*1.9,
		sub = knife(params[0].zw,params[1].xy,sideX,scaleCorrect);//current stroke
	if (params[0].x < 1e8) sub = max(0.,sub-knife(params[0].xy,params[0].zw,sideX,scaleCorrect));//subtract last stroke
	val -= sub*params[1].w;
	gl_FragColor = vec4(max(val,0.),0,0,1);
}
</script>


<script type="text/plain" id="FloodFillClear.glsl">
//flood fill/clear pumpkin
attribute vec3 vpos;
varying vec2 uv;
uniform sampler2D map;
uniform vec4 params[2];/*1.xy = current point
1.z = fill/clear value
1.w = opacity*/

void VERT() {
	uv = vpos.xy*.5+.5;
	gl_Position = vec4(vpos,1);
}
void FRAG() {
	vec4 val = texture2D(map,uv), ov;
	if (val.y == 1.) val.x = params[1].z;
	if (length(params[1].xy-uv) < 1./512.) val.y = 1.;
	
	#define AS(ox,oy) ov = texture2D(map,uv+vec2(ox,oy)/vec2(1024,512)); if (ov.y == 1. && abs(ov.x-val.x) < params[1].w) {val.y = 1.;}
	
	AS(-1,-1);
	AS(0,-1);
	AS(1,-1);
	AS(1,0);
	AS(1,1);
	AS(0,1);
	AS(-1,1);
	AS(-1,0);
	
	gl_FragColor = val;
}
</script>


<script type="text/plain" id="FloodReset.glsl">
//reset flood buffer state
attribute vec3 vpos;
varying vec2 uv;
uniform sampler2D map;

void VERT() {
	uv = vpos.xy*.5+.5;
	gl_Position = vec4(vpos,1);
}
void FRAG() {
	vec4 val = texture2D(map,uv), ov;
	val.y = 0.;
	gl_FragColor = val;
}
</script>


<script type="text/plain" id="Copy.glsl">
//copy and flip y for saving
attribute vec3 vpos;
varying vec2 uv;
uniform sampler2D buffer;

void VERT() {
	uv = vpos.xy*vec2(.5,-.5)+.5;
	gl_Position = vec4(vpos,1);
}
void FRAG() {
	gl_FragColor = texture2D(buffer,uv);
}
</script>
